#include <float.h>
#include <libavcodec/avcodec.h>
#include <libavutil/internal.h>
#include "a52dec.h"
#include "audio_out_internal.h"
#include "ffmpeg_stub.h"

#define MAX_FRAME_SIZE      4096
#define CHANNELS            2

ao_instance_t * ao_raw32_open (void);

typedef struct {
  AVClass *class;
  double drc_scale;
  double gain;
  a52dec_t dec;
  AVFrame frame;
} priv_t;

static av_cold int decode_init(AVCodecContext *dc)
{
  priv_t *p=dc->priv_data;

  dc->sample_fmt=AV_SAMPLE_FMT_S32;
  dc->channels=CHANNELS;

  __avcodec_get_frame_defaults(&p->frame);
  dc->coded_frame=&p->frame;

  if (NULL==a52dec_open(&p->dec,ao_raw32_open))
	return -1;

  p->dec.disable_dynrng=(0.0==p->drc_scale);

  if (0.0<p->gain)
    p->dec.gain=p->gain;

  return 0;
}

static int decode_frame(AVCodecContext *dc, void *outdata, int *got_frame,
    AVPacket *avpkt)
{
  priv_t *p=dc->priv_data;
  AVFrame *frame=outdata;
  raw32_instance_t *output=(raw32_instance_t *)p->dec.output;
  int ret;

  p->frame.nb_samples=MAX_FRAME_SIZE;

  if ((ret=dc->get_buffer(dc,&p->frame))<0) {
    __av_log(dc,AV_LOG_ERROR,"get_buffer() failed\n");
    return ret;
  }

  output->int32_samples=(int32_t *)p->frame.data[0];
  output->max_size=p->frame.linesize[0];
  output->offs=0;
  output->size=0;

  a52dec_decode_data(&p->dec,avpkt->data,avpkt->data+avpkt->size);

  dc->bit_rate=p->dec.bit_rate;
  dc->sample_rate=p->dec.sample_rate;
  p->frame.nb_samples=output->size/sizeof(int32_t)/CHANNELS;
  *(AVFrame *)frame=p->frame;
  *got_frame=1;

  return avpkt->size;
}

static av_cold int decode_close(AVCodecContext *dc)
{
  priv_t *p=dc->priv_data;

  a52dec_close(&p->dec);

  return 0;
}

static void decode_flush(AVCodecContext *dc)
{
  priv_t *p=dc->priv_data;

  a52dec_flush(&p->dec);
}

static const char* item_name(void *ctx)
{
  return __av_default_item_name(ctx);
}

#define OFFSET(x) offsetof(priv_t, x)
#define PAR (AV_OPT_FLAG_DECODING_PARAM | AV_OPT_FLAG_AUDIO_PARAM)
static const AVOption options[] = {
  { "drc_scale","dynamic range compression",OFFSET(drc_scale),
      AV_OPT_TYPE_DOUBLE,{.dbl=1.0},0.0,1.0,PAR },
  { "gain","apply gain",OFFSET(gain),
      AV_OPT_TYPE_DOUBLE,{.dbl=1.0},0.0,DBL_MAX,PAR },
  { NULL},
};

static AVClass liba52_decoder_class = {
  .class_name = "LibA52 decoder",
  .item_name  = item_name,
  .option     = options,
  .version    = LIBAVUTIL_VERSION_INT,
};

AVCodec liba52_adapter={
  .name="liba52",
  .type=AVMEDIA_TYPE_AUDIO,
  .id=CODEC_ID_AC3,
  .priv_data_size=sizeof(priv_t),
  .init=decode_init,
  .decode=decode_frame,
  .close=decode_close,
  .flush=decode_flush,
  .capabilities=CODEC_CAP_DR1,
  .long_name=NULL_IF_CONFIG_SMALL("ATSC A/52A (AC-3)"),
  .sample_fmts=(const enum AVSampleFormat[]){
    AV_SAMPLE_FMT_S32,
    AV_SAMPLE_FMT_NONE
  },
  .priv_class=&liba52_decoder_class
};

void liba52_adapter_register()
{
  a52dec_start();
  __avcodec_register(&liba52_adapter);
}
