/*
 * libmad_adapter.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <float.h>
#include <stdio.h>
#include <libavutil/internal.h>
#include "ffmpeg_stub.h"

# include "madutil.h"
# include "libmad_stub.h"
# include "ffmpeg_stub.h"

# ifdef fprintf
# undef fprintf
# endif

# define MPEG_BUFSZ	40000	/* 2.5 s at 128 kbps; 1 s at 320 kbps */
# define MIN(a,b) ((a)<(b)?(a):(b))

#define MAX_FRAME_SIZE      4096
#define CHANNELS            2

typedef struct {
  AVClass *class;
  struct mad_decoder decoder;
  mad_sync_t sync;
  double gain;

  struct input {
    AVPacket *avpkt;

    unsigned char *data;
    unsigned long length;

    int eof;
  } input;

  struct output {
    unsigned int bitrate, samplerate, nchannels, nsamples;
    int32_t *data;
	int data_size;
  } output;

  AVFrame frame;
} priv_t;

/*
 * NAME:	decode->input_read()
 * DESCRIPTION:	(re)fill decoder input buffer by reading a file descriptor
 */
static
enum mad_flow decode_input_read(void *data, struct mad_stream *stream)
{
  priv_t *ctx=data;
  struct input *input = &ctx->input;
  int len;

  if (input->eof)
    return MAD_FLOW_STOP;

  if (stream->next_frame) {
    memmove(input->data, stream->next_frame,
	    input->length = &input->data[input->length] - stream->next_frame);
  }

  len=MIN(MPEG_BUFSZ-input->length-MAD_BUFFER_GUARD,input->avpkt->size);
  memcpy(input->data + input->length, input->avpkt->data, len);

  if (len == 0) {
    input->eof = 1;

    assert(MPEG_BUFSZ - input->length >= MAD_BUFFER_GUARD);

    while (len < MAD_BUFFER_GUARD)
      input->data[input->length + len++] = 0;
  }

  __mad_stream_buffer(stream, input->data, input->length += len);

  return MAD_FLOW_CONTINUE;
}

static inline
signed int scale(mad_fixed_t sample)
{
  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  return sample << 3;
}

/*
 * NAME:	decode->output()
 * DESCRIPTION: configure audio module and output decoded samples
 */
static
enum mad_flow decode_output(void *data, struct mad_header const *header,
			    struct mad_pcm *pcm)
{
  priv_t *ctx=data;
  struct output *output=&ctx->output;
  unsigned int nchannels, nsamples;
  mad_fixed_t const *left_ch, *right_ch;

  /* pcm->samplerate contains the sampling frequency */
  nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  output->bitrate=header->bitrate;
  output->samplerate=pcm->samplerate;
  output->nchannels=nchannels;
  output->nsamples+=nsamples;

  if (1.0==ctx->gain) {
    while (nsamples--) {
	  *output->data++=scale(*left_ch++);
      output->data_size+=sizeof *output->data;

      if (nchannels == 2) {
        *output->data++=scale(*right_ch++);
        output->data_size+=sizeof *output->data;
	  }
    }
  }
  else {
	mad_fixed_t gain=mad_f_tofixed(ctx->gain);

    while (nsamples--) {
	  *output->data++=scale(mad_f_mul(gain,*left_ch++));
      output->data_size+=sizeof *output->data;

      if (nchannels == 2) {
        *output->data++=scale(mad_f_mul(gain,*right_ch++));
        output->data_size+=sizeof *output->data;
	  }
    }
  }

  return MAD_FLOW_CONTINUE;
}

void libmad_adapter_set_gain(AVCodecContext *dc, double gain)
{
  ((priv_t *)dc->priv_data)->gain=gain;
}

static av_cold int decode_init(AVCodecContext * dc)
{
  priv_t *p=dc->priv_data;

  __mad_decoder_init(&p->decoder, p,
		   decode_input_read,
		   0, 0,
		   decode_output,
		   0, 0);
  //p->gain=1.0;
  p->input.data=__av_malloc(MPEG_BUFSZ);
  p->input.length=0;

  dc->sample_fmt=AV_SAMPLE_FMT_S32;
  dc->channels=CHANNELS;

  __avcodec_get_frame_defaults(&p->frame);
  dc->coded_frame=&p->frame;

  p->decoder.sync=__av_malloc(sizeof(*p->decoder.sync));
  mad_sync_init(&p->sync, &p->decoder);

  return 0;
}

static av_cold int decode_close(AVCodecContext *dc)
{
  priv_t *p=dc->priv_data;
  mad_sync_finish(&p->sync);
  __av_free(p->decoder.sync);
  __av_free(p->input.data);
  p->decoder.sync=NULL;
  __mad_decoder_finish(&p->decoder);

  return 0;
}

static int decode_frame(AVCodecContext * dc,
                        void *data, int *data_size,
                        AVPacket *avpkt)
{
  priv_t *p=dc->priv_data;
  mad_sync_t *sync=&p->sync;
  int ret;

  if (MAD_SYNC_STATE_OUTER_LOOP!=sync->state)
    return -1;

  p->frame.nb_samples=MAX_FRAME_SIZE;

  if ((ret=dc->get_buffer(dc,&p->frame))<0) {
    __av_log(dc, AV_LOG_ERROR, "get_buffer() failed\n");
    return ret;
  }

  p->input.avpkt=avpkt;
  p->output.data=(int32_t *)p->frame.data[0];
  //p->output->max_size=p->frame.linesize[0];
  p->output.data_size=0;

  do {
    mad_sync_next(sync);
  } while (MAD_SYNC_STATE_INNER_LOOP==sync->state);

  dc->bit_rate=p->output.bitrate;
  dc->sample_fmt=AV_SAMPLE_FMT_S32;
  dc->sample_rate=p->output.samplerate;
  dc->channels=0==p->output.nchannels?CHANNELS:p->output.nchannels;

  p->frame.nb_samples=p->output.data_size/sizeof(int32_t)/dc->channels;
  *(AVFrame *)data=p->frame;
  *data_size=1; // got frame

  if (p->input.eof&&MAD_SYNC_STATE_OUTER_LOOP==sync->state)
    mad_sync_next(sync);

  return avpkt->size;
}

static const char* item_name(void *ctx)
{
  return __av_default_item_name(ctx);
}

#define OFFSET(x) offsetof(priv_t, x)
#define PAR (AV_OPT_FLAG_DECODING_PARAM | AV_OPT_FLAG_AUDIO_PARAM)
static const AVOption options[] = {
  { "gain","apply gain",OFFSET(gain),
      AV_OPT_TYPE_DOUBLE,{.dbl=1.0},0.0,DBL_MAX,PAR },
  { NULL},
};

static AVClass libmad_decoder_class = {
  .class_name = "LibMAD decoder",
  .item_name  = item_name,
  .option     = options,
  .version    = LIBAVUTIL_VERSION_INT,
};

AVCodec libmad_adapter =
{
  .name=LIBMAD,
  .type=AVMEDIA_TYPE_AUDIO,
  .id=CODEC_ID_MP3,
  .priv_data_size=sizeof(priv_t),
  .init=decode_init,
  .close=decode_close,
  .decode=decode_frame,
  .capabilities=CODEC_CAP_DR1,
  .long_name=NULL_IF_CONFIG_SMALL("MP3 (MPEG audio layer 3)"),
  .sample_fmts=(const enum AVSampleFormat[]){
    AV_SAMPLE_FMT_S32,
    AV_SAMPLE_FMT_NONE
  },
  .priv_class=&libmad_decoder_class
};

void libmad_adapter_register()
{
  __avcodec_register(&libmad_adapter);
}
